"""
Commands related to file management and transfer:
- push
- pull
- clean
"""

from os import getcwd
from os.path import isfile, join
from typing import Optional

import click
from termcolor import colored

from .config import Config
from .user_interaction import print_info, print_warning, prompt_user_yes_no
from .utils import run_local_cmd, run_remote_command_inria
from .cli_utils import option_force


########
# Push #############################################################################################
########

def push_(
    rsync_destination: str,
    rsync_e_opt: str = None,
    last_push_fn_suffix: str = '',
    force: bool = False
) -> None:
    """
    Sync the project folder to the remote workstation.

    Args:
        force (bool):   Run the sync command even if no local changes were detected.
    """
    config: Config = Config()

    rsync_cmd: list[str] = [
        'rsync',
        '-zravut',
        '--human-readable',
        '--executability',
        '--delete'
    ]

    if isfile(config.exclude_file):
        rsync_cmd.append(f"--exclude-from='{config.exclude_file}'")

    # Source
    rsync_cmd.append('.')

    # Smart push
    # The 'smart push' mechanism stores the folder state (list of files along with their timestamps
    # and size) at each push. Then we are able to check if something has changed locally
    # since last push and run the rsync push command only if necessary.
    # This avoids opening a useless ssh tunnel with the server (which can sometimes last a few
    # seconds).
    last_push_file_path: str = f'.remi/last_push{last_push_fn_suffix}.log'
    last_push_file_path_temp: str = f'.remi/.last_push_temp{last_push_fn_suffix}.log'

    def _get_last_push_log(filename) -> str:

        _last_push_log: str = ''
        if isfile(filename):
            with open(filename, 'r') as last_push_log_file:
                _last_push_log = last_push_log_file.read()

        return _last_push_log

    rsync_check_cmd: list[str] = rsync_cmd + ['--list-only', '>', last_push_file_path_temp]

    # Run rsync check command
    run_local_cmd(command=rsync_check_cmd, check=True)

    if _get_last_push_log(last_push_file_path) == _get_last_push_log(last_push_file_path_temp) \
            and not force:
        print_info('Nothing has changed since last push. Cancelling push.')
        return

    # Eventually, add remote shell option (for proxy-jumping)
    if rsync_e_opt:
        rsync_cmd += ['-e', f"'{rsync_e_opt}'"]

    # Else, something has changed: so we sync with remote.
    rsync_cmd.append(rsync_destination)

    # Run the rsync command
    run_local_cmd(command=rsync_cmd, check=True)

    # Now that we know the sync was succesfull, store the sync state permanently
    run_local_cmd(['mv', last_push_file_path_temp, last_push_file_path])


def push_inria(force: bool = False) -> None:
    config: Config = Config()

    # Not pushing if already working from Inria.
    if config.working_from_inria and getcwd() == config.project_remote_path:
        print_warning(
            "You seem to be working directly from your Inria desktop and the `project_remote_path`"
            " defined in the config file is the same as the current working directory."
            f"\n=> Nothing will get 'pushed' and the current working directory (`{getcwd()}`) will"
            " be used by the cluster and other Inria desktops to access the project files."
        )
        return

    host_str: str = colored(config.desktop.ip_adress, 'cyan')
    print_info(
        f"Pushing project folder to remote host `{host_str}`",
        bold=True
    )

    rsync_destination: str
    rsync_e_opt: str = ''

    if config.working_from_inria:
        # Destination
        rsync_destination = config.project_remote_path

    else:
        # Bastion
        if config.bastion.enable:
            rsync_e_opt = f"ssh -J {config.username}@{config.bastion.hostname}"

        # Destination
        rsync_destination = \
            f"{config.username}@{config.desktop.ip_adress}:{config.project_remote_path}"

    push_(
        rsync_destination=rsync_destination,
        rsync_e_opt=rsync_e_opt,
        force=force
    )


option_push_force = option_force(
    help_message="Push the project even though no local changes were noticed."
)


@click.command()
@option_push_force
def push(force: bool) -> None:
    """
    Sync the project folder to the remote workstation.

    Args:
        force (bool):   Run the sync command even if no local changes were detected.
    """
    push_inria(force=force)


########
# Pull #############################################################################################
########

# pylint: disable=unused-argument
def _callback_argument_output_path_default(ctx, param, value) -> str:
    if value == '':
        return Config().output_path

    return value


argument_pull_remote_path: click.argument = click.argument(
    'remote_path',
    default='',
    type=str,
    callback=_callback_argument_output_path_default,
    nargs=1
)

option_pull_force: click.option = option_force(
    help_message='Do not prompt for any confirmation.'
)

def pull_(
    remote_path: str,
    source_prefix: str,
    rsync_e_opt: str = None,
    force: bool = False
) -> None:

    is_dir: bool = remote_path.endswith('/')
    node_type: str = 'directory' if is_dir else 'file'

    if not force:
        if not is_dir:
            print_info('Assuming that the provided path targets a FILE.')
            print_warning(
                f"==> If you meant to target a DIRECTORY, add a `/` at the end --> `{remote_path}/`"
            )

    remote_path_str: str = colored(remote_path, 'cyan')
    print_info(
        f"Pulling {node_type} `{remote_path_str}` from the remote host",
        bold=True
    )

    rsync_cmd: list[str] = [
        'rsync',
        '-zravut',
        '--human-readable',
        '--executability'
    ]

    # Eventually, add remote shell option (for proxy-jumping)
    if rsync_e_opt:
        rsync_cmd += ['-e', f"'{rsync_e_opt}'"]

    # Source
    rsync_cmd.append(join(source_prefix, remote_path))

    # Destination (copy at the same path on the local computer)
    rsync_cmd.append(remote_path)

    run_local_cmd(command=rsync_cmd, check=True)


@click.command()
@argument_pull_remote_path
@option_pull_force
def pull(remote_path: str,
         force: bool) -> None:
    """
    Sync the content of the provided remote directories to the equivalent folder to the local
    machine.

    Args:
        remote_path (str):  The remote folder to sync to the local computer.
        force (bool):       Do not ask for a confirmation before pulling.
                                Eventually conflicting local files might be overridden.
    """
    config: Config = Config()
    source_prefix: str = config.project_remote_path

    rsync_e_opt: str = ''

    # Not pushing if already working from Inria.
    if config.working_from_inria:
        if getcwd() == config.project_remote_path:
            print_warning(
                "You seem to be working directly from your Inria desktop and the"
                " `project_remote_path` defined in the config file is the same as the current"
                " working directory."
                "\n=> Nothing will get 'pulled'."
            )
            return

    else:
        # Bastion
        if config.bastion.enable:
            rsync_e_opt = f"ssh -J {config.username}@{config.bastion.hostname}"

        # Source
        source_prefix = f'{config.username}@{config.desktop.ip_adress}:' + source_prefix

    pull_(
        remote_path=remote_path,
        source_prefix=source_prefix,
        rsync_e_opt=rsync_e_opt,
        force=bool
    )


#########
# Clean ############################################################################################
#########

option_clean_force: click.option = option_force(
    help_message="Do not ask for confirmation before deleting the folder content."
)
argument_clean_directory: click.argument = click.argument(
    'directory',
    default='',
    type=str,
    callback=_callback_argument_output_path_default,
    nargs=1
)

def clean_(directory: str,
           force: bool,
           remote_path: str) -> Optional[list[str]]:

    directory_str: str = directory
    if not directory_str.endswith('/'):
        directory_str += '/'
    directory_str = colored(directory_str, 'cyan')
    if force or prompt_user_yes_no(
            question=f"Do you really want to delete the content of the `{directory_str}` directory"
                     " on the remote location ?"
    ):
        print_info(
            text=f"Cleaning folder `{directory_str}` on the remote location.",
            bold=True
        )

        clean_cmd: list[str] = ['rm', '-rf', join(remote_path, directory, '*')]

        return clean_cmd

    print_info('Aborting.')


@click.command()
@option_clean_force
@argument_clean_directory
def clean(directory: str, force: bool) -> None:
    """
    Clean the content of a folder on the remote computer.

    Args:
        directory (str):        The directory of which content should be deleted.
        force (bool):           Do not ask for a confirmation before cleaning.
    """
    clean_cmd: Optional[list[str]] = clean_(
        directory=directory,
        remote_path=Config().project_remote_path,
        force=force
    )
    if clean_cmd:
        run_remote_command_inria(clean_cmd, check=False)



"""
A class representing a remote server (an application that run on the remote and is accessible via
http).
A ssh tunnel transports the traffic on the selected port between the local and remote computers.
"""

import sys
from abc import ABC
from signal import signal, SIGINT
from subprocess import Popen
from socket import gethostname
from typing import Optional
from termcolor import colored
import click

from ..user_interaction import to_bold, print_info, print_error
from ..config import Config
from ..utils import run_remote_command_inria
from ..file_transfer import push_inria
from .cli_utils import option_port
from ..cli_utils import option_hostname


def _print_stdout(text: str) -> None:
    print(colored('[STDOUT]', 'blue', attrs=['bold']), text)


def _print_stderr(text: str) -> None:
    print(colored('[STDERR]', 'magenta', attrs=['bold']), text)


class SshTunnel:
    """
    Class for running and stopping a SSH tunnel.
    """

    def __init__(
        self,
        hostname: str,
        port: int
    ) -> None:

        self.hostname: str = hostname
        self.port: int = port

        self.process: Optional[Popen] = None

    def start(self) -> None:
        """
        Start the ssh tunnel.
        """
        config: Config = Config()

        print_info(f'Running SSH port forwarding for port {self.port}')
        print_info(to_bold('on host: ') + str(self.hostname))

        ssh_tunnel_cmd: list[str] = ['ssh', '-N']

        ssh_tunnel_cmd += ["-L", f"{self.port}:localhost:{self.port}"]

        # Bastion ProxyJump
        if config.bastion.enable:
            ssh_tunnel_cmd += ['-J', f'{config.username}@{config.bastion.hostname}']

        ssh_tunnel_cmd.append(f'{config.username}@{self.hostname}')

        # Run the command
        print_info(to_bold('SSH tunnel command: ') + ' '.join(ssh_tunnel_cmd))

        # pylint: disable=consider-using-with
        self.process = Popen(args=ssh_tunnel_cmd)

    def stop(self) -> None:
        """
        Stop the ssh tunnel.
        """
        if self.process is not None:
            print_info('Killing the ssh tunnel process')
            self.process.kill()


class RemoteServer(ABC):
    """
    Abstract class for remote servers.
    """

    def __init__(
        self,
        server_start_command: str,
        server_stop_command: str,
        port: int,
        hostname: str,
        open_browser: bool,
        local_url: str
    ) -> None:

        self.server_start_command: list[str] = self._wrap_command(server_start_command)
        self.server_stop_command: list[str] = self._wrap_command(server_stop_command)
        self.port = port
        self.hostname = hostname

        self.open_browser: bool = open_browser
        self.local_url: str = local_url

        self.ssh_tunnel: Optional[SshTunnel] = None
        if gethostname() not in self.hostname:
            self.ssh_tunnel = SshTunnel(
                hostname=hostname,
                port=port
            )

        self.server_process: Popen = None

    def _server_start_callback(self, server_output_line: str) -> bool:
        """
        Given an line from the server standard output tells is this line means that the server is
        properly running.
        """
        raise NotImplementedError

    @staticmethod
    def _wrap_command(command: str) -> None:

        config: Config = Config()

        # cd to project directory
        wrapped_command: list[str] = ['cd', config.project_remote_path, '&&']

        # Server command
        wrapped_command += command.split()

        return wrapped_command

    def _stop(self) -> None:

        self.stop_server()

        # Kill the ssh tunnel process.
        if self.ssh_tunnel is not None:
            self.ssh_tunnel.stop()

        # Finally, quit remi.
        sys.exit(0)

    # pylint: disable=unused-argument
    def _sigint_handler(self, signal_received, frame) -> None:
        """
        See more here (https://docs.python.org/3/library/signal.html#signal.signal).
        This function implements a signal handler
        """
        self._stop()

    def _standby_loop(self) -> None:
        """
        A loop that runs as long as the server is running and displays its output.
        """
        # Check if the server launch command have not failed (due to port busy).
        # In case it is running, wait to have the url (which contains the token).
        return_code: int = None
        server_stdout_line: str = ''
        while True:
            return_code = self.server_process.poll()

            # Case 1: the server is running
            if return_code is None:
                server_stdout_line = self.server_process.stdout.readline().strip()

                print(server_stdout_line)

                if self._server_start_callback(server_stdout_line):
                    break

            # Case 2: command failed
            # Example: a server was already running on this port
            elif return_code > 0:

                print_error(
                    f"{self.__class__.__name__} command failed with exit code {return_code}"
                )

                server_stdout_lines: list[str] = self.server_process.stdout.readlines()

                # Print stderr output
                for line in server_stdout_lines:
                    print(line.strip())

                # Kill the ssh tunnel process before leaving
                if self.ssh_tunnel is not None:
                    self.ssh_tunnel.stop()

                # Exit remi with a non-zero exit code
                sys.exit(1)

            else:
                print_info(f"{self.__class__.__name__} has exited succesfully.")

                # Kill the ssh tunnel process before leaving
                if self.ssh_tunnel is not None:
                    self.ssh_tunnel.stop()

                # Exit remi
                sys.exit(0)

    def start(self, callback: callable = None) -> None:
        """
        Run the server on the remote computer.
        """
        # First, sync the project
        push_inria()

        # Run the server

        # Server process
        self.server_process: Popen = run_remote_command_inria(
            command=self.server_start_command,
            hostname=self.hostname,
            background=True
        )

        # Wait for the server to succesfully start.
        self._standby_loop()

        print_info('The server is now succesfully running !')

        # Then, open an ssh tunnel in the background (if working remotely)
        if gethostname() not in self.hostname:
            self.ssh_tunnel.start()

        # Catch Ctrl-c signal from local user and use it as a remote server killer.
        signal(SIGINT, self._sigint_handler)

        # If enabled, open the browser on the local machine.
        browser_cmd: str = Config().remote_servers.browser_cmd
        if self.open_browser and browser_cmd != '':
            print_info('Opening the service in the browser')

            # pylint: disable=consider-using-with
            Popen(args=browser_cmd.split() + [self.local_url])

        print_info('To quit, simply press `Ctrl+c`')

        self._standby_loop()

    def stop_server(self) -> None:
        run_remote_command_inria(
            command=self.server_stop_command,
            hostname=self.hostname,
            check=False
        )


@click.command('ssh-tunnel')
@option_hostname
@option_port(service_name='ssh tunnel')
def ssh_bridge(hostname: str, port: int) -> None:
    """
    Run a standalone ssh tunnel.
    """
    ssh_tunnel: SshTunnel = SshTunnel(
        hostname=hostname,
        port=port
    )
    ssh_tunnel.start()

    def _sigint_handler(signal_received, frame) -> None:
        ssh_tunnel.stop()

    signal(SIGINT, _sigint_handler)

    # Block indefinitely
    ssh_tunnel.process.wait()



"""
Jupyter notebook support.
"""

import click

from .remote_server import RemoteServer
from ..cli_utils import OrderedGroup, option_hostname
from .cli_utils import option_port, option_open_browser


class Jupyter(RemoteServer):
    """
    Jupyter remote server.
    """

    def __init__(
            self,
            port: int,
            hostname: str,
            open_browser: bool = True
    ) -> None:

        super().__init__(
            server_start_command="python -m jupyter notebook --no-browser"
                                 f" --port={port} --port-retries=0",
            server_stop_command=f"python -m jupyter notebook stop {port}",
            port=port,
            hostname=hostname,
            open_browser=open_browser,
            local_url=''
        )

    def _server_start_callback(self, server_output_line: str) -> bool:

        if server_output_line.startswith('http://localhost:'):
            self.local_url = server_output_line
            return True

        return False


_option_jupyter_port = option_port(service_name='jupyter')


@click.group(cls=OrderedGroup, invoke_without_command=True)
@click.pass_context
@_option_jupyter_port
@option_hostname
@option_open_browser
def jupyter(
        ctx,
        port: int,
        hostname: str,
        open_browser: bool
) -> None:
    """
    Run a jupyter server on the remote host.
    """
    if ctx.invoked_subcommand is None:
        ctx.invoke(
            jupyter_run,
            port=port,
            hostname=hostname,
            open_browser=open_browser
        )


@jupyter.command('run')
@_option_jupyter_port
@option_hostname
@option_open_browser
def jupyter_run(
        port: int,
        hostname: str,
        open_browser: bool
) -> None:
    """
    Run a jupyter server on the remote host.
    """
    Jupyter(
        port=port,
        hostname=hostname,
        open_browser=open_browser
    ).start()


@jupyter.command('stop')
@_option_jupyter_port
@option_hostname
def jupyter_stop(port: int, hostname: str) -> None:
    """
    Stop the jupyter server on the remote host.
    """
    Jupyter(
        port=port,
        hostname=hostname
    ).stop_server()



"""
Utility functions for interacting with oarsub.
"""

from datetime import datetime
from os.path import join

from ...config import Config, Oarsub
from ...user_interaction import to_bold, print_info

def print_cluster_request_prop(
        property_name: str,
        value: str
) -> None:

    num_dots: int = 25 - len(property_name)
    to_print: str = '\t' + to_bold(property_name + ':') + '.' * num_dots + str(value)
    print_info(to_print)


def _print_cluster_request(
        oarsub_config: Oarsub,
        host_id: str,
        job_name: str
) -> None:

    print_info(to_bold('Cluster request:'))
    print_cluster_request_prop(
        property_name='Job name',
        value=job_name
    )
    print_cluster_request_prop(
        property_name='Number of hosts',
        value=oarsub_config.num_hosts
    )
    if oarsub_config.num_cpu_cores > 0:
        print_cluster_request_prop(
            property_name='Number of CPU cores',
            value=oarsub_config.num_cpu_cores
        )
    print_cluster_request_prop(
        property_name='Number of GPUs',
        value=oarsub_config.num_gpus
    )
    print_cluster_request_prop(
        property_name='Walltime',
        value=oarsub_config.walltime
    )
    print_cluster_request_prop(
        property_name='Cluster name',
        value=oarsub_config.cluster_name
    )
    if host_id != '':
        print_cluster_request_prop(
            property_name='Host ID',
            value=host_id
        )
    if oarsub_config.custom_property_query:
        print_cluster_request_prop(
            property_name='Custom property query',
            value=oarsub_config.custom_property_query
        )
    print_cluster_request_prop(
        property_name='Besteffort',
        value=oarsub_config.besteffort
    )
    print_cluster_request_prop(
        property_name='Idempotent',
        value=oarsub_config.idempotent
    )


def get_oarsub_cmd_prefix(
        host_id: str,
        job_name: str
) -> list[str]:

    # Get the oarsub config
    oarsub_config: Oarsub = Config().oarsub

    # Oarsub command
    oarsub_cmd: list[str] = ['oarsub']

    # Name the job
    oarsub_cmd += ['--name', job_name]

    # Resource list
    resource_list: str = f'/host={oarsub_config.num_hosts}'
    if oarsub_config.num_cpu_cores > 0:
        resource_list += f'/core={oarsub_config.num_cpu_cores}'

    if oarsub_config.num_gpus > 0:
        resource_list += f'/gpudevice={oarsub_config.num_gpus}'

    if oarsub_config.walltime != '':
        resource_list += f',walltime={oarsub_config.walltime}'

    oarsub_cmd += ['-l', resource_list]

    # Check that the cluster name is valid
    assert oarsub_config.cluster_name in [
        'beagle',
        'cp',
        'kinovis',
        'mistis',
        'nanod',
        'perception',
        'thoth'
    ], f"Unknown cluster name: {oarsub_config.cluster_name}"

    # Property list
    property_list: str
    # Using directly the provided property query, ignoring other options.
    if oarsub_config.custom_property_query:
        property_list = '\"' + oarsub_config.custom_property_query + '\"'

    # Using the other config settings.
    else:
        property_list = f"\"cluster='{oarsub_config.cluster_name}'"

        if host_id:
            assert 4 <= len(host_id) <= 5
            assert host_id.startswith('gpu') \
                   or host_id.startswith('node')
            cluster_host_name: str = f'{host_id}-{oarsub_config.cluster_name}.inrialpes.fr'

            property_list += f" AND host='{cluster_host_name}'"

        property_list += '\"'
    oarsub_cmd += ['-p', property_list]

    if oarsub_config.besteffort:
        oarsub_cmd += ['-t', 'besteffort']

    if oarsub_config.idempotent:
        oarsub_cmd += ['-t', 'idempotent']

    _print_cluster_request(
        oarsub_config=oarsub_config,
        job_name=job_name,
        host_id=host_id
    )

    return oarsub_cmd


def add_standard_output_params(
    oarsub_log_path: str,
    base_name: str = ''
) -> list[str]:

    time_stamp: str = datetime.now().strftime('%Y-%m-%d_%H-%M-%S')

    if not base_name:
        base_name = f"OAR.{time_stamp}.%jobid%.%jobname%"

    return [
        '--stdout', join(oarsub_log_path, f'{base_name}.stdout'),
        '--stderr', join(oarsub_log_path, f'{base_name}.stderr')
    ]



"""
Run code on the GriCAD cluster.
"""

import sys
import subprocess
from typing import Union, Optional
from os.path import join

from termcolor import colored
import click

from ...cli_utils import OrderedGroup, option_no_push
from ..cli_utils import option_script, argument_command
from .cli_utils import argument_oar_job_id, argument_oar_job_id_list
from ...utils import run_remote_command
from ...user_interaction import print_info, to_bold, print_error
from ... import config
from ...file_transfer import (
    push_,
    option_push_force,
    pull_,
    argument_pull_remote_path,
    option_pull_force,
    clean_,
    argument_clean_directory,
    option_clean_force
)
from ..singularity import get_binding_sub_command
from .oarsub import print_cluster_request_prop, add_standard_output_params


def _get_ssh_proxy_command(cluster_name: str = None) -> list[str]:

    g_config: config.Gricad = config.Config().gricad
    if not cluster_name:
        cluster_name = g_config.prefered_cluster

    ssh_proxy_command: str = (
        f"ProxyCommand ssh -q {g_config.username}@access-gricad.univ-grenoble-alpes.fr"
        f" nc -w 60 {cluster_name} 22"
    )

    return ssh_proxy_command


def _run_remote_command_gricad(
    command: list[str],
    cluster_name: str = None,
    check: bool = False,
    force_tty_allocation: bool = False
) -> Union[subprocess.Popen, int]:

    g_config: config.Gricad = config.Config().gricad

    if not cluster_name:
        cluster_name = g_config.prefered_cluster

    hostname_args: list[str] = [
        '-o', _get_ssh_proxy_command(cluster_name=cluster_name),
        f'{g_config.username}@{cluster_name}'
    ]

    return run_remote_command(
        command=command,
        hostname=hostname_args,
        pretty_hostname=cluster_name,
        check=check,
        force_tty_allocation=force_tty_allocation,
    )


# pylint: disable=unused-argument
def _callback_option_gricad_job_name(ctx, param, value) -> str:
    if not value:
        return config.Config().gricad.oarsub.job_name

    return value


_option_gricad_job_name = click.option(
    '-n', '--job-name',
    default='',
    callback=_callback_option_gricad_job_name,
    help='Custom job name'
)

# pylint: disable=unused-argument
def _callback_option_gricad_gpu_model(ctx, param, value) -> str:
    if not value:
        value = config.Config().gricad.oarsub.gpu_model

    # Give more flexibility w.r.t the input: 'a100' --> 'A100'
    value = value.upper()

    if value not in ['A100', 'V100', 'T4']:
        print_error(f"Invalid `gpu_model` '{value}'."
                    " Supported values are: 'A100', 'V100' and 'T4'.")
        sys.exit(1)

    return value


_option_gricad_gpu_model = click.option(
    '-g', '--gpu-model',
    callback=_callback_option_gricad_gpu_model,
    help="GPU model ('A100', 'V100' or 'T4')"
)

@click.group(cls=OrderedGroup, invoke_without_command=True)
@option_script
@_option_gricad_gpu_model
@_option_gricad_job_name
@option_no_push
@click.pass_context
def gricad(
    ctx,
    script_name: str,
    job_name: str,
    gpu_model: str,
    no_push: bool
) -> None:
    """
    Run a script on the cluster (see config for options).
    """
    if ctx.invoked_subcommand is None:
        ctx.invoke(
            cluster_script,
            script_name=script_name,
            job_name=job_name,
            gpu_model=gpu_model,
            no_push=no_push
        )


#########
# Setup ############################################################################################
#########

@gricad.command()
def setup() -> None:
    """
    Set up remote project on the Gricad cluster.
    """
    _config: config.Config = config.Config()
    g_config: config.Gricad = _config.gricad

    print_info(text=f'Setting up gricad project: {_config.project_name}',
               bold=True)

    paths_to_create: str = g_config.project_remote_path
    paths_to_create += '/{' + ','.join(
        [
            _config.output_path,
            _config.notebooks_path,
            _config.oarsub_log_path
        ]
    ) + '}'

    # Create the project directory on the remote and the output folder within it.
    _run_remote_command_gricad(
        command=['mkdir', '-p', paths_to_create],
        check=True
    )
    _push_gricad()

    print_info('Setup complete !')

########
# Push #############################################################################################
########

def _push_gricad(force: bool = False):

    g_config: config.Gricad = config.Config().gricad

    cluster_name: str = g_config.prefered_cluster

    print_info("Pushing project folder to gricad", bold=True)

    rsync_destination: str = f"{g_config.username}@{cluster_name}:{g_config.project_remote_path}"
    rsync_e_opt: str = f'ssh -o "{_get_ssh_proxy_command(cluster_name=cluster_name)}"'

    push_(
        rsync_destination=rsync_destination,
        rsync_e_opt=rsync_e_opt,
        last_push_fn_suffix='_gricad',
        force=force
    )


@gricad.command()
@option_push_force
def push(force: bool) -> None:
    """
    Sync the project folder to the gricad storage.

    Args:
        force (bool):   Run the sync command even if no local changes were detected.
    """
    _push_gricad(force=force)


########
# Pull #############################################################################################
########

@gricad.command()
@argument_pull_remote_path
@option_pull_force
def pull(remote_path: str, force: bool) -> None:
    """
    Sync the content of the provided remote directories to the equivalent folder to the local
    machine.

    Args:
        remote_path (str):  The remote folder to sync to the local computer.
        force (bool):       Do not ask for a confirmation before pulling.
                                Eventually conflicting local files might be overridden.
    """
    g_config: config.Gricad = config.Config().gricad

    rsync_e_opt: str = f'ssh -o "{_get_ssh_proxy_command()}"'

    # Source
    source_prefix: str = \
        f'{g_config.username}@{g_config.prefered_cluster}:{g_config.project_remote_path}'

    pull_(
        remote_path=remote_path,
        source_prefix=source_prefix,
        rsync_e_opt=rsync_e_opt,
        force=bool
    )


#########
# Clean ############################################################################################
#########

@gricad.command()
@option_clean_force
@argument_clean_directory
def clean(directory: str, force: bool) -> None:
    """
    Clean the content of a folder on the remote computer.

    Args:
        directory (str):        The directory of which content should be deleted.
        force (bool):           Do not ask for a confirmation before cleaning.
    """
    clean_cmd: Optional[list[str]] = clean_(
        directory=directory,
        remote_path=config.Config().gricad.project_remote_path,
        force=force
    )
    if clean_cmd:
        _run_remote_command_gricad(clean_cmd, check=False)


###########
# Cluster ##########################################################################################
###########


def _print_cluster_request(
    oarsub_config: config.GricadOarsub,
    job_name: str,
    cluster_name: str,
    gpu_model: str
) -> None:

    print_info(to_bold('Cluster request:'))
    print_cluster_request_prop(
        property_name='Gricad cluster',
        value=cluster_name
    )
    print_cluster_request_prop(
        property_name='Job name',
        value=job_name
    )
    print_cluster_request_prop(
        property_name='Number of nodes',
        value=oarsub_config.num_nodes
    )
    if oarsub_config.num_cpus > 0:
        print_cluster_request_prop(
            property_name='Number of CPUs',
            value=oarsub_config.num_cpus
        )
    if oarsub_config.num_cpu_cores > 0:
        print_cluster_request_prop(
            property_name='Number of CPU cores',
            value=oarsub_config.num_cpu_cores
        )

    print_cluster_request_prop(
        property_name='Number of GPUs',
        value=oarsub_config.num_gpus
    )
    if oarsub_config.num_gpus > 0 and gpu_model:
        print_cluster_request_prop(
            property_name='GPU model',
            value=gpu_model
        )

    print_cluster_request_prop(
        property_name='Walltime',
        value=oarsub_config.walltime
    )


def get_oarsub_cmd_prefix(
    job_name: str,
    cluster_name: str,
    gpu_model: str
) -> list[str]:

    # Get the oarsub config
    g_config: config.Gricad = config.Config().gricad
    oarsub_config: config.GricadOarsub = g_config.oarsub

    # Oarsub command
    oarsub_cmd: list[str] = ['oarsub']

    # Name the job
    oarsub_cmd += ['--name', job_name]

    # Resource list
    resource_list: str = f'/nodes={oarsub_config.num_nodes}'
    if oarsub_config.num_cpus > 0:
        resource_list += f'/cpu={oarsub_config.num_cpus}'
    if oarsub_config.num_cpu_cores > 0:
        resource_list += f'/core={oarsub_config.num_cpu_cores}'

    if oarsub_config.num_gpus > 0:
        resource_list += f'/gpu={oarsub_config.num_gpus}'

    if oarsub_config.walltime != '':
        resource_list += f',walltime={oarsub_config.walltime}'

    oarsub_cmd += ['-l', resource_list]

    if oarsub_config.num_gpus > 0 and gpu_model:
        oarsub_cmd += ['-p', f'"gpumodel=\'{gpu_model}\'"']

    oarsub_cmd += ['--project', g_config.project_name]

    _print_cluster_request(
        oarsub_config=oarsub_config,
        job_name=job_name,
        cluster_name=cluster_name,
        gpu_model=gpu_model
    )

    return oarsub_cmd


def _cluster(
    sub_command: str,
    script_name: str = '',
    command: tuple[str] = '',
    gpu_model: str = '',
    job_name: str = '',
    no_push: bool = False
) -> None:

    assert sub_command in ['script', 'interactive', 'command']

    # Sync project.
    if not no_push:
        _push_gricad()
        print()

    _config: config.Config = config.Config()
    g_config: config.Gricad = _config.gricad

    cluster_name: str = g_config.prefered_cluster

    oarsub_cmd: list[str] = get_oarsub_cmd_prefix(
        job_name=job_name,
        cluster_name=cluster_name,
        gpu_model=gpu_model
    )

    # Interactive
    if sub_command == 'interactive':
        print_info('Connecting to cluster in interactive mode',
                   bold=True)
        oarsub_cmd += ['--interactive']

    # Script or command
    else:
        oarsub_cmd += add_standard_output_params(
            oarsub_log_path=join(
                g_config.project_remote_path,
                _config.oarsub_log_path
            ),
            base_name=g_config.oarsub.log_file_name
        )

        job_cmd: list[str] = ['cd', g_config.project_remote_path, '&&']

        use_container: bool = g_config.use_container

        if use_container:

            job_cmd += ['/usr/local/bin/singularity', 'exec']

            job_cmd += get_binding_sub_command(
                project_remote_path=g_config.project_remote_path
            )

            # Enable GPU support and provide the path to the container image
            job_cmd += ['--nv', g_config.singularity_image]

        if sub_command == 'script':
            script_name_pretty: str = colored(script_name, 'yellow')
            print_info(
                to_bold('Running script ')
                + colored(script_name_pretty, 'yellow')
                + ' on the cluster'
            )
            print()

            job_cmd += ['bash', script_name]

        elif sub_command == 'command':
            command_merge: str = ' '.join(command)
            command_merge_pretty: str = colored(command_merge, 'yellow')
            print_info(
                to_bold('Running command ')
                + colored(command_merge_pretty, 'yellow')
                + ' on the cluster'
            )
            print()

            job_cmd += command

        if _config.verbose:
            job_cmd_pretty: str = colored(' '.join(job_cmd), 'cyan')
            print_info(f'Command to run on the cluster: {job_cmd_pretty}')

        oarsub_cmd += ['\"' + ' '.join(job_cmd) + '\"']

    _run_remote_command_gricad(
        command=oarsub_cmd,
        cluster_name=cluster_name,
        check=True,
        force_tty_allocation=True
    )

@gricad.command('script')
@option_script
@_option_gricad_gpu_model
@_option_gricad_job_name
@option_no_push
def cluster_script(
    script_name: str,
    gpu_model: str,
    job_name: str,
    no_push: bool
) -> None:
    _cluster(
        sub_command='script',
        script_name=script_name,
        gpu_model=gpu_model,
        job_name=job_name,
        no_push=no_push
    )


@gricad.command('command')
@argument_command
@_option_gricad_gpu_model
@_option_gricad_job_name
@option_no_push
def cluster_command(
    command: list[str],
    job_name: str,
    gpu_model: str,
    no_push: bool
) -> None:

    _cluster(
        sub_command='command',
        command=command,
        gpu_model=gpu_model,
        job_name=job_name,
        no_push=no_push
    )


@gricad.command('interactive')
@_option_gricad_gpu_model
@_option_gricad_job_name
@option_no_push
def cluster_interactive(
    job_name: str,
    gpu_model: str,
    no_push: bool
) -> None:
    _cluster(
        sub_command='interactive',
        job_name=job_name,
        gpu_model=gpu_model,
        no_push=no_push
    )


@gricad.command('chandler')
def cluster_chandler() -> None:
    """
    Run `chandler` on the cluster to list compute nodes occupation (free/busy).
    """
    _run_remote_command_gricad(
        command="chandler",
        force_tty_allocation=True,
        check=True
    )

@gricad.command('recap')
def cluster_recap() -> None:
    """
    Run `recap.py` on the cluster to list compute nodes information (CPU, GPU…).
    """
    _run_remote_command_gricad(
        command="recap.py",
        force_tty_allocation=True,
        check=True
    )

@gricad.command('stat')
def cluster_stat() -> None:
    """
    Get some information about your running/planned jobs thanks to `oarstat`.
    """
    _run_remote_command_gricad(
        command=f"oarstat -u {config.Config().gricad.username}",
        force_tty_allocation=True,
        check=True
    )


@gricad.command('connect')
@argument_oar_job_id
def cluster_connect(oar_job_id: int) -> None:
    """
    Connect to a job on the cluster.

    Arguments:
        OAR_JOB_ID: The id of the job you want to kill (It has to have been run by you)'
    """
    _run_remote_command_gricad(
        command=f"oarsub -C {oar_job_id}",
        force_tty_allocation=True,
        check=True
    )


@gricad.command('kill')
@argument_oar_job_id_list
def gricad_kill(oar_job_ids: tuple[int]) -> None:
    """
    Kill a job on the gricad cluster.

    Arguments:
        OAR_JOB_IDS: The ids of the jobs you want to kill (they have to have been run by you)'
    """
    oar_job_ids_str: str = ' '.join(
        str(job_id)
        for job_id in oar_job_ids
    )
    _run_remote_command_gricad(
        command=f"oardel {oar_job_ids_str}",
        check=True
    )



"""
Default configuration.
"""
from os.path import join, isdir, isfile, basename
from os import mkdir, getcwd
from typing import Any
from socket import gethostname

from ..user_interaction import prompt_user_yes_no, prompt_user_value, print_info, print_warning

def _config_file_template(
        project_name: str,
        username: str,
        pc_name: str,
        project_remote_path: str,
        gricad_username: str = 'GRICAD_USERNAME',
        gricad_prefered_cluster: str = 'bigfoot',
        gricad_project_name: str = 'GRICAD_PROJECT_NAME',
        gricad_project_remote_path: str = 'GRICAD_PROJECT_REMOTE_PATH'
) -> str:
    return f"""# If true, makes remi's output more explicit
verbose: false

# Name for your project
project_name: {project_name}


# Inria username
username: {username}


# Location of the project on the remote computer
project_remote_path: {project_remote_path}


desktop:
  # Name of your Inria workstation
  hostname: {pc_name}

  ip_adress: {pc_name}.inrialpes.fr

  # Whether to use the singularity container when running jobs on the workstations.
  use_container: true

  # Desktop background jobs
  background:
    # Which backend to use (`screen` or `tmux`)
    backend: screen

    # Whether to keep the session alive after the job has ended.
    # It lets you attach to the session to see the program output.
    # If 'false', the session will be closed when the job is over and stdout/stderr will be
    # lost.
    # CAUTION: If true, you will have to manually re-attach and close the session.
    keep_session_alive: false


# Bastion used to ssh into Inria resources
bastion:
  enable: true
  hostname: bastion.inrialpes.fr
  username: {username}


# Singularity container options
singularity:
  # The name of the 'recipe' file (`.def`) to build the singularity container.
  def_file_name: container.def

  # The name of the singularity image.
  output_sif_name: container.sif

  # A dictionnary of binds for the singularity container.
  # If the value is empty (''), the mount point is the same as the path on the host.
  # By default, the project folder is bound within the singularity container: This configuration
  # then allows you to add extra locations.
  # Example:
  #     /path_on_host/my_data: /path_in_container/my_data
  bindings:

  # Whether to bind beegfs. (It will be available as `/beegfs/` in the container).
  bind_beegfs: false

  # The HOMEDIR is mounted by default by singularity.
  # If you want to disable this behavior, set the following option to true.
  # Learn more: https://docs.sylabs.io/guides/3.1/user-guide/bind_paths_and_mounts.html#using-no-home-and-containall-flags
  no_home: false


# Oarsub options (for more details on `oarsub`, please refer to
# https://oar.imag.fr/docs/latest/user/commands/oarsub.html).
oarsub:

  # Job name
  job_name: {project_name}

  # Number of hosts requested.
  num_hosts: 1

  # Number of cpu cores requested.
  # If the value is 0, all the cores for the requested cpus will be used.
  num_cpu_cores: 0

  # Number of GPUs requested.
  # If the value is 0, no GPU will be requested (CPU only).
  num_gpus: 1

  # The maximum allowed duration for your job.
  walltime: '72:00:00'

  # The name of the requested cluster (perception, mistis, thoth...)
  cluster_name: perception

  # Optionnaly specify the id of a specific node (gpu3, node2...)
  host_id:

  # If the options above are too restricive for your use-case, you may
  # directly provide a property list that will be provided to `oarsub` with the
  # `-p` flag.
  custom_property_query:

  # Whether to schedule the job in the besteffort queue.
  besteffort: true

  # Whether to set the job as idempotent (see oarsub documentation for more details).
  idempotent: false

  # Template name for the log files.
  # By default the log files are named: YYYY-MM-DD_hh-mm-ss.JOB_ID.JOB_NAME.std[out, err]
  # Ex: 2022-06-12_14-47-52.7502202.bce_type2_EaConv1d.stdout
  #
  # You can use '%jobid%' and '%jobname%' to reference the job id and name.
  # '.stdout' and '.stdout' is appended at the end automatically.
  #
  # Example:
  #     log_file_name: 'oar_log_%jobid%_%jobname%'
  log_file_name:


gricad:
  username: {gricad_username}

  # The Gricad cluster (bigfoot, dahu, luke, froggy)
  prefered_cluster: {gricad_prefered_cluster}

  # The Gricad project you are a member of
  # see: https://gricad-doc.univ-grenoble-alpes.fr/en/services/perseus-ng/3_project/
  project_name: {gricad_project_name}

  # Location of the project on the remote computer
  project_remote_path: {gricad_project_remote_path}

  # Whether to use the singularity container when running jobs on Gricad
  use_container: true

  # The name of the singularity image.
  singularity_image: container.sif

  oarsub:

    # Job name
    job_name: {project_name}

    # Number of nodes requested.
    num_nodes: 1

    # Number of cpus requested (per requested node).
    # If the value is 0, all the cpus for the requested node will be used.
    num_cpus: 0

    # Number of cpu cores requested.
    # If the value is 0, all the cores for the requested cpus will be used.
    num_cpu_cores: 0

    # Number of GPUs requested.
    # If the value is 0, no GPU will be requested (CPU only).
    num_gpus: 1

    # GPU model (leave blank if you have no preference)
    # Possible values: 'A100', 'V100', 'T4'
    gpu_model: V100

    # The maximum allowed duration for your job.
    walltime: '48:00:00'

    # Template name for the log files.
    # By default the log files are named: YYYY-MM-DD_hh-mm-ss.JOB_ID.JOB_NAME.std[out, err]
    # Ex: 2022-06-12_14-47-52.7502202.bce_type2_EaConv1d.stdout
    #
    # You can use '%jobid%' and '%jobname%' to reference the job id and name.
    # '.stdout' and '.stdout' is appended at the end automatically.
    #
    # Example:
    #     log_file_name: 'oar_log_%jobid%_%jobname%'
    log_file_name:


# Remote servers
# Remote servers are applications that run on a remote computer and can be accessed from your local
# browser thanks to remi.
#
# Two such servers are supported right now:
# - Jupyter notebook
# - TensorBoard
remote_servers:
  # The command to run for opening the local browser (`<browser_cmd> <url>`)
  browser_cmd: firefox

  # Jupyter notebook
  jupyter:
    # The port (local and remote) for the server
    port: 8080

    # If true, automatically open the jupyter notebook in the local browser.
    open_browser: true

  # TensorBoard
  tensorboard:
    # The port (local and remote) for TensorBoard
    port: 9090

    # Directory from where to run tensorboard.
    logdir: 'output/'

    # If true, automatically open TensorBoard in the local browser.
    open_browser: true
"""


def create_default_project():
    """
    Generate the configuration file and the exclude file.
    """
    dot_remi: str = '.remi/'
    config_file: str = join(dot_remi, 'config.yaml')

    if not isfile(config_file):
        print_info("Generating configuration file")
        cwd: str = basename(getcwd())
        project_name: str = prompt_user_value(
            question='Project_name',
            default_value=cwd
        )
        username: str = prompt_user_value(question="Inria username")
        pc_name: str = prompt_user_value(question="Inria hostname (name of your desktop machine)")
        if gethostname() == pc_name:
            project_remote_path = getcwd()
        else:
            project_remote_path = '/'.join(
                [
                    '/scratch',
                    pc_name,
                    username,
                    '.remi_projects',
                    project_name
                ]
            )

        config_values: dict[str, Any] = {
            'project_name': project_name,
            'username': username,
            'project_remote_path': project_remote_path,
            'pc_name': pc_name,
        }

        # Gricad
        if prompt_user_yes_no(question="Configure Gricad ?"):
            config_values['gricad_username']: str = prompt_user_value(question='Perseus username')

            config_values['gricad_project_name']: str = prompt_user_value(
                question='Perseus project',
                default_value='pr-ml3ri'
            )
            config_values['gricad_project_remote_path']: str = \
                f"/bettik/{config_values['gricad_username']}/.remi_projects/{project_name}"

            config_values['gricad_prefered_cluster']: str = prompt_user_value(
                question='Gricad cluster',
                default_value='bigfoot'
            )

        config_string: str = _config_file_template(**config_values)

    # Once the user has provided a config, create the `.remi/` directory
    if not isdir(dot_remi):
        print(f"Creating {dot_remi}")
        mkdir(dot_remi)

    # Then, dump the config to the config file if it doesn't already exists
    if not isfile(config_file):
        with open(config_file, 'w') as file_stream:
            file_stream.write(config_string)

    else:
        print_warning(f'Config file (`{config_file}`) already exists: skipping.')

    exclude_file = join(dot_remi, 'exclude.txt')

    # Create a .gitignore in `.remi/` to exclude everything but the config and exclude files.
    dot_gitignore_path: str = join(dot_remi, '.gitignore')
    if not isfile(dot_gitignore_path):
        with open(dot_gitignore_path, 'w') as dot_gitignore:
            dot_gitignore.write('*\n')
            dot_gitignore.write('!config.yaml\n')
            dot_gitignore.write('!exclude.txt\n')

    if not isfile(exclude_file):
        print_info('Creating boilerplate exclude file')
        with open(str(exclude_file), 'w') as file_stream:
            file_stream.write(
                '\n'.join(
                    [
                        '.remi',
                        'output/',
                        'notebooks/',
                        '.git',
                        '.envrc',
                        '__pycache__',
                        '.ipynb_checkpoints',
                        'logs',
                        '.DS_Store',
                        '.*.swp',
                        '*.egg-info/',
                        '**/__pycache__/'
                        '.idea',
                        '.mypy_cache/',
                        'venv/',
                        '*.sif',
                        'build-temp-*',
                        '.pytest_cache/'
                    ]
                )
            )
    else:
        print_warning(f'Exclude file (`{exclude_file}`) already exists: skipping.')

    print_info("We have created a standard configurations file and exclude list specifying files"
               " and folders that shouldn't be copied to server.")
    print_info("You can edit them at `.remi/config.yaml` and `.remi/exclude.txt`")



"""
Functions specific to the Inria cluster.
"""

from os.path import join
import subprocess
from typing import Union

import click
from termcolor import colored

from ...config import Config
from ...user_interaction import to_bold, print_info
from ...file_transfer import push_inria
from ...utils import run_remote_command_inria
from ...cli_utils import OrderedGroup, option_no_push, option_xforwarding
from ..cli_utils import option_script, option_no_build, argument_command
from .cli_utils import argument_oar_job_id, argument_oar_job_id_list

from ..singularity import get_binding_sub_command, build_singularity_container
from .oarsub import get_oarsub_cmd_prefix, add_standard_output_params


def _run_remote_command_cluster(
    command: list[str],
    enable_x_forwarding: bool = False,
) -> Union[subprocess.Popen, int]:

    return run_remote_command_inria(
        command=command,
        hostname='access2-cp.inrialpes.fr',
        check=True,
        force_tty_allocation=True,
        enable_x_forwarding=enable_x_forwarding,
        use_bastion=True
    )


def _cluster(
    sub_command: str,
    script_name: str = '',
    command: tuple[str] = '',
    host_id: str = '',
    job_name: str = '',
    enable_x_forwarding: bool = False,
    no_build: bool = False,
    no_push: bool = False
) -> None:

    assert sub_command in ['script', 'interactive', 'command']

    # Sync project.
    if not no_push:
        push_inria()
        print()

    config: Config = Config()

    oarsub_cmd: list[str] = get_oarsub_cmd_prefix(
        job_name=job_name,
        host_id=host_id
    )

    # Interactive
    if sub_command == 'interactive':
        print_info('Connecting to cluster in interactive mode',
                   bold=True)
        oarsub_cmd += ['--interactive']

    # Script or command
    else:

        # X11 xforwarding makes no sense when not in interactive mode.
        enable_x_forwarding = False

        if not no_build:
            print()
            build_singularity_container()
            print()

        oarsub_cmd += add_standard_output_params(
            oarsub_log_path=join(
                config.project_remote_path,
                config.oarsub_log_path
            ),
            base_name=config.oarsub.log_file_name
        )

        job_cmd: list[str] = ['cd', config.project_remote_path, '&&', 'singularity', 'exec']

        # Path bindings between host system and singularity container
        job_cmd += get_binding_sub_command(
            project_remote_path=config.project_remote_path,
            bindings=config.singularity.bindings,
            no_home=config.singularity.no_home
        )

        if config.singularity.bind_beegfs:
            job_cmd += ['--bind', f'/services/scratch/perception/{config.username}/:/beegfs/']

        # Export the PYTHONPATH environment variable so as to use the project python module
        # job_cmd += ['--env', 'PYTHONPATH=.']

        if config.oarsub.num_gpus > 0:
            job_cmd.append('--nv')

        job_cmd.append(config.singularity.output_sif_name)

        if sub_command == 'script':
            script_name_pretty: str = colored(script_name, 'yellow')
            print_info(
                to_bold('Running script ')
                + colored(script_name_pretty, 'yellow')
                + ' on the cluster'
            )
            print()

            job_cmd += ['bash', script_name]

        elif sub_command == 'command':
            command_merge: str = ' '.join(command)
            command_merge_pretty: str = colored(command_merge, 'yellow')
            print_info(
                to_bold('Running command ')
                + colored(command_merge_pretty, 'yellow')
                + ' on the cluster'
            )
            print()

            job_cmd += command

        if config.verbose:
            job_cmd_pretty: str = colored(' '.join(job_cmd), 'cyan')
            print_info(f'Command to run on the cluster: {job_cmd_pretty}')

        oarsub_cmd += ['\"' + ' '.join(job_cmd) + '\"']

    _run_remote_command_cluster(
        command=oarsub_cmd,
        enable_x_forwarding=enable_x_forwarding,
    )


#######
# CLI #
#######
# pylint: disable=unused-argument
def _callback_option_cluster_job_name(ctx, param, value) -> str:
    if not value:
        return Config().oarsub.job_name

    return value


_option_cluster_job_name = click.option(
    '-n', '--job-name',
    default='',
    callback=_callback_option_cluster_job_name,
    help='Custom job name'
)

# pylint: disable=unused-argument
def _callback_option_host_id(ctx, param, value) -> str:
    if not value:
        return Config().oarsub.host_id

    return value


_option_host_id = click.option(
    '-h', '--host',
    'host_id',
    default='',
    callback=_callback_option_host_id,
    type=str,
    help='A specific host id.'
)


@click.group(cls=OrderedGroup, invoke_without_command=True)
@option_script
@_option_host_id
@_option_cluster_job_name
@option_no_push
@option_no_build
@click.pass_context
def cluster(
    ctx,
    script_name: str,
    job_name: str,
    host_id: str,
    no_push: bool,
    no_build: bool
) -> None:
    """
    Run a script on the cluster (see config for options).
    """
    if ctx.invoked_subcommand is None:
        ctx.invoke(
            cluster_script,
            script_name=script_name,
            job_name=job_name,
            host_id=host_id,
            no_push=no_push,
            no_build=no_build
        )


@cluster.command('script')
@option_script
@_option_host_id
@_option_cluster_job_name
@option_no_push
@option_no_build
def cluster_script(
    script_name: str,
    host_id: str,
    job_name: str,
    no_push: bool,
    no_build
) -> None:

    _cluster(
        sub_command='script',
        script_name=script_name,
        host_id=host_id,
        job_name=job_name,
        no_push=no_push,
        no_build=no_build
    )


@cluster.command('command')
@argument_command
@_option_host_id
@_option_cluster_job_name
@option_no_push
@option_no_build
def cluster_command(
    command: list[str],
    host_id: str,
    job_name: str,
    no_push: bool,
    no_build: bool
) -> None:

    _cluster(
        sub_command='command',
        command=command,
        host_id=host_id,
        job_name=job_name,
        no_push=no_push,
        no_build=no_build
    )


@cluster.command('interactive')
@_option_host_id
@_option_cluster_job_name
@option_xforwarding
@option_no_push
def cluster_interactive(
    host_id: str,
    job_name: str,
    x_forwarding: bool,
    no_push: bool
) -> None:

    _cluster(
        sub_command='interactive',
        host_id=host_id,
        job_name=job_name,
        enable_x_forwarding=x_forwarding,
        no_push=no_push
    )


@cluster.command('stat')
@click.option(
    '-t', '--team',
    default=False,
    is_flag=True,
    help='Query the current jobs for the RobotLearn team members'
)
def cluster_stat(
    team: bool
) -> None:
    """
    Get the state of your jobs.
    """
    command: list[str] = ['oarstat']

    if team:
        command += [
            '|', 'grep',
            "'" + '\\|'.join(
                (
                    'aauterna',
                    'aballou',
                    'creinke',
                    'galepage',
                    'lairale',
                    'visanche',
                    'wguo',
                    'xilin',
                    'xbie',
                    'yixu'
                )
            ) + "'"
        ]
    else:
        command += ['-u', Config().username]

    _run_remote_command_cluster(
        command=command
    )


@cluster.command('connect')
@argument_oar_job_id
def cluster_connect(oar_job_id: int) -> None:
    """
    Connect to a job on the cluster.

    Arguments:
        OAR_JOB_ID: The id of the job you want to kill (It has to have been run by you)'
    """
    _run_remote_command_cluster(
        command=f"oarsub -C {oar_job_id}",
    )


@cluster.command('kill')
@argument_oar_job_id_list
def cluster_kill(oar_job_ids: tuple[int]) -> None:
    """
    Kill a job on the cluster.

    Arguments:
        OAR_JOB_IDS: The ids of the jobs you want to kill (they have to have been run by you)'
    """
    oar_job_ids_str: str = ' '.join(
        str(job_id)
        for job_id in oar_job_ids
    )
    _run_remote_command_cluster(
        command=f"oardel {oar_job_ids_str}",
    )



"""
Common options for remote compute commands
"""

import sys
from os.path import isfile
import click

from ..user_interaction import print_error


# pylint: disable=unused-argument
def _callback_option_script(ctx, param, value) -> str:
    # Make sure the script is a valid file
    if ctx.invoked_subcommand not in (None, 'script'):
        if not isfile(value):
            print_error(f'Script `{value}` does not exist.')
            sys.exit(1)

    return value


# pylint: disable=unused-argument
def _callback_argument_command(ctx, param, value) -> str:
    if len(value) == 0:
        print_error('No command provided. Exiting')
        sys.exit(1)

    return value


option_interactive = click.option(
    '-i', '--interactive',
    is_flag=True,
    help='Open an interactive session instead of running the script'
)

option_script = click.option(
    '-s', '--script',
    'script_name',
    default='script.sh',
    type=str,
    callback=_callback_option_script,
    help='Path to a bash script to run'
)

option_background = click.option(
    '-b', '--background',
    default=False,
    is_flag=True,
    help='Run the script in the background'
)

option_no_build = click.option(
    '--no-build',
    default=False,
    is_flag=True,
    help='Prevent trying to build the singularity container.'
)

argument_command = click.argument(
    'command',
    nargs=-1,
    type=str,
    callback=_callback_argument_command
)



"""
Commands for setting up remi for a project.
"""

import click

from .config import Config
from .config.defaults import create_default_project
from .user_interaction import print_info, prompt_user_yes_no
from .utils import create_remote_directory, run_local_cmd
from .cli_utils import option_hostname
from .file_transfer import push_inria


@click.command()
def init() -> None:
    """
    Initialize the project in the current working directory. Generate the configuration files.
    """
    print_info(
        text="Creating a brand new remi project.",
        bold=True
    )
    create_default_project()


@click.command()
@option_hostname
def setup(hostname: str) -> None:
    """
    Set up remote project location.

    Args:
        hostname (str):     The hostname of the remote computer.
    """
    config: Config = Config()

    print_info(
        text=f'Setting up remote for project: {config.project_name}',
        bold=True
    )

    paths_to_create: str = config.project_remote_path
    paths_to_create += '/{' + ','.join(
        [
            config.output_path,
            config.notebooks_path,
            config.oarsub_log_path
        ]
    ) + '}'

    # Create the project directory on the remote and the output folder within it.
    create_remote_directory(
        path=paths_to_create,
        hostname=hostname
    )
    push_inria()

    print_info('Setup complete !')


@click.command()
def update() -> None:
    """
    Update remi to the latest version
    """
    if prompt_user_yes_no("Do you wish to update 'remi' to the latest version ?"):
        run_local_cmd(
            command=[
                'pip',
                'install',
                '-U',
                'remote-inria'
            ],
            check=True
        )
    else:
        print_info("All right the, aborting.")



"""
Functions to interact with singularity.
"""

from typing import Optional
import subprocess
from os.path import isfile
from termcolor import colored

import click

from ..cli_utils import option_force
from ..user_interaction import to_bold, print_info, print_warning
from ..utils import run_remote_command_inria, create_remote_directory
from ..file_transfer import push_inria
from ..config import Config


def get_binding_sub_command(
    project_remote_path: str,
    bindings: dict[str, Optional[str]] = None,
    no_home: bool = False,
) -> list[str]:
    """
    Path bindings between host system and singularity container
    """
    binding_sub_cmd: list[str] = []

    if bindings is not None:
        if project_remote_path not in bindings:
            binding_sub_cmd += ['--bind', project_remote_path]
        for host_path, container_path in bindings.items():
            binding_sub_cmd += ['--bind', host_path]
            if container_path:
                binding_sub_cmd[-1] += ':' + container_path
    else:
        binding_sub_cmd += ['--bind', project_remote_path]

    if no_home:
        binding_sub_cmd.append('--no-home')

    return binding_sub_cmd


def build_singularity_container(force: bool = False) -> None:

    singularity_config = Config().singularity

    def_file_name: str = singularity_config.def_file_name
    output_sif_name: str = singularity_config.output_sif_name

    if not isfile(def_file_name):
        print_warning(f'Recipe file not found (`{def_file_name}`). Aborting')
        return

    # Smart build
    # The 'smart build' mechanism stores the last built recipe file at each build
    # under `.remi/last_built_recipe.def`.
    # Then we are able to check if something has changed locally since last build and run the
    # singularity command only if necessary (if `container.def` differs from
    # `.remi/last_built_recipe.def`).
    # This avoids opening a useless ssh tunnel with the server (which can sometimes last a few
    # seconds) and asking a confirmation to the user.
    last_built_recipe_path: str = '.remi/last_built_recipe.def'

    last_built_recipe: str = ''
    if isfile(last_built_recipe_path):
        with open(last_built_recipe_path, 'r') as last_recipe_built_file:
            last_built_recipe = last_recipe_built_file.read()

    with open(def_file_name, 'r') as current_recipe_file:
        current_recipe: str = current_recipe_file.read()

    # Check if something has changed
    if current_recipe == last_built_recipe and not force:
        print_info('Nothing has changed since last build. Cancelling build.')
        return

    # Else, something has changed: so we build the image remotely.

    # Run singularity build on Inria remote
    build_command: list[str] = [
        'cd', Config().project_remote_path, '&&',
        'sudo', 'singularity', 'build'
    ]

    print_info(
        to_bold('Building container image ') + '`' + colored(output_sif_name, 'yellow') + '`'
        + to_bold(' from recipe ') + '`' + colored(def_file_name, 'yellow') + '`'
    )

    if force:
        build_command += ['--force']

    build_command += [output_sif_name, def_file_name]

    return_code: int = run_remote_command_inria(
        command=build_command,
        check=False,
        force_tty_allocation=True
    )

    if return_code == 0:
        print_info("Build succesfull.")
        # If the build command succeeded, save the current recipe as the 'last built recipe'
        subprocess.run(
            ['cp', '-f', def_file_name, last_built_recipe_path],
            check=True
        )
    else:
        print_info('You canceled the build remotely')


@click.command('build-container')
@option_force(help_message='Build the container even if the sif image already exists.')
def build_container(force: bool) -> None:
    """
    Build the singularity container on the remote desktop.
    """
    # First, sync project with remote
    push_inria()

    build_singularity_container(force=force)



"""
Common options for remi commands
"""

from collections import OrderedDict
import click

from .config import Config

# Common options, arguments and their eventual callbacks


# pylint: disable=unused-argument
def _callback_option_hostname(ctx, param, value) -> str:

    config: Config = Config()
    # If no hostname was provided, use the one from the config file.
    if value == '':
        value = config.desktop.ip_adress

    # Make sure the provided hostname is a valid Inria hostname
    elif not value.endswith('.inrialpes.fr'):
        value += '.inrialpes.fr'

    return value


option_hostname: click.option = click.option(
    '-h', '--hostname',
    default='',
    type=str,
    callback=_callback_option_hostname,
    help='Name of an Inria computer'
)
option_no_push: click.option = click.option(
    '--no-push',
    default=False,
    is_flag=True,
    help='Prevent syncing the project files.'
)

option_xforwarding = click.option(
    '-X', '--xforwarding',
    'x_forwarding',
    default=False,
    is_flag=True,
    help='Enables X forwarding in SSH.'
)

def option_force(
        help_message: str,
        default: bool = False
):
    return click.option(
        '-f', '--force',
        default=default,
        is_flag=True,
        help=help_message
    )


class OrderedGroup(click.Group):
    """
    Override of the click.Group class in order to preserve custom ordering of commands in `--help`.
    """

    def __init__(
            self,
            name=None,
            commands=None,
            **attrs
    ) -> None:

        super().__init__(name, commands, **attrs)
        #: the registered subcommands by their exported names.
        self.commands = commands or OrderedDict()

    def list_commands(self, ctx):
        """
        Override of the `list_commands` method.
        """
        return self.commands



"""
Functions to interact with the user through the command line.
"""

from typing import Any

from termcolor import colored

def to_bold(text: str):
    return colored(text, attrs=['bold'])


def print_info(
        text: str,
        newline: bool = False,
        bold: bool = False,
        end: str = '\n'
) -> None:

    prefix: str = ''
    if newline:
        prefix += '\n'
    prefix += colored('[INFO]', 'green', attrs=['bold'])

    if bold:
        text = to_bold(text)

    print(prefix, text, end=end)


def print_warning(
        text: str,
        newline: bool = False,
        bold: bool = False,
        end: str = '\n'
) -> None:

    prefix: str = ''
    if newline:
        prefix += '\n'
    prefix += colored('[WARNING]', 'yellow', attrs=['bold'])

    if bold:
        text = to_bold(text)

    print(prefix, text, end=end)


def print_error(
        text: str,
        newline: bool = False,
        bold: bool = False,
        end: str = '\n'
) -> None:

    prefix: str = ''
    if newline:
        prefix += '\n'
    prefix += colored('[ERROR]', 'red', attrs=['bold'])

    if bold:
        text = to_bold(text)

    print(prefix, text, end=end)


def prompt_user_yes_no(question: str) -> bool:
    answer: str = ''
    while answer not in ['y', 'n']:
        print_warning(question, end='')
        answer = input(" [y/n] ").lower()

    return answer == 'y'

def prompt_user_value(
        question: str,
        default_value: Any = None
) -> Any:

    question: str = question

    if default_value is not None:
        question += f" [{colored(default_value, 'yellow')}]"
    question += ': '

    return input(question).strip() or default_value



"""
Run code on a remote desktop computer.
"""

from socket import gethostname
import click

from ..cli_utils import (
    OrderedGroup,
    option_hostname,
    option_xforwarding,
    option_no_push
)
from .cli_utils import (
    option_script,
    option_background,
    option_no_build,
    argument_command
)

from ..config import Config
from ..user_interaction import to_bold, print_info, print_warning, colored
from ..utils import run_remote_command_inria
from ..file_transfer import push_inria
from .singularity import build_singularity_container, get_binding_sub_command


def _desktop(
    sub_command: str,
    hostname: str,
    script_name: str = '',
    background: bool = False,
    session_name: str = '',
    attach: bool = False,
    command: tuple[str] = '',
    enable_x_forwarding: bool = False,
    no_push: bool = False,
    no_build: bool = False
) -> None:

    assert sub_command in ['script', 'interactive', 'command']

    # Sync project.
    if not no_push:
        push_inria()

    config: Config = Config()

    cmd: list[str] = ['cd', config.project_remote_path, '&&']

    singularity_cmd: list[str] = []

    use_container: bool = config.desktop.use_container

    if use_container:
        if not no_build:
            build_singularity_container()
            print()

        singularity_cmd = ['singularity']
        if sub_command == 'interactive':
            singularity_cmd.append('shell')
        else:
            singularity_cmd.append('exec')

        singularity_cmd += get_binding_sub_command(
            project_remote_path=config.project_remote_path,
            bindings=config.singularity.bindings,
            no_home=config.singularity.no_home
        )

        if config.singularity.bind_beegfs:
            singularity_cmd += ['--bind', f'/mnt/beegfs/perception/{config.username}/:/beegfs/']

        # Export the PYTHONPATH environment variable so as to use the project python module
        singularity_cmd += ['--env', 'PYTHONPATH=.']

        # Enable GPU support and provide the path to the container image
        singularity_cmd += ['--nv', config.singularity.output_sif_name]

    # Interactive
    if sub_command == 'interactive':
        if gethostname() in hostname:
            # TODO, if the current working directory is different than the `project_remote_path`
            # (which is the case by default), we should perform necessary stuff to work from the
            # project directory, with env/singularity...
            print_warning(f"You seem to be already working directly from {hostname}."
                          "\n=> An interactive session does not make sense. Aborting.")
            return

        print_info(
            f'Connecting to remote desktop ({hostname}) in interactive mode.',
            bold=True
        )

        if use_container:
            cmd += singularity_cmd

        # No container
        else:
            cmd += ['exec', 'bash']

    # Script or command
    else:
        cmd_suffix: str = ''

        # Script
        if sub_command == 'script':
            print_info(
                to_bold('Running script ')
                + colored(script_name, 'yellow')
                + ' on remote desktop '
                + colored(hostname, 'magenta')
            )
            print()

            cmd_suffix += f"bash {script_name}"

        # Command
        else:
            assert len(command) > 0

            command_merge: str = ' '.join(command)
            print_info(
                to_bold('Running command ')
                + colored(command_merge, 'yellow')
                + ' on remote desktop '
                + colored(hostname, 'magenta')
            )
            print()

            cmd_suffix += command_merge

        # Use `screen` to run a command in the background.
        if background:
            if config.desktop.background.backend == 'screen':
                cmd += ['screen', '-dmS']
            elif config.desktop.background.backend == 'tmux':
                cmd += ['tmux', 'new-session', '-d']

                # Set the session working directory to the project remote path
                cmd += ['-c', config.project_remote_path]

                # Name the session
                cmd += ['-s']

            cmd += [session_name, 'bash', '-c']

            # If singularity + container, add singularity prefix in the cmd_suffix.
            if use_container:
                cmd_suffix = ' '.join(singularity_cmd) + ' ' + cmd_suffix

            # Keep session alive
            if config.desktop.background.keep_session_alive:
                print_warning(
                    f"The `{config.desktop.background.backend}` session will be kept alive once the"
                    " job command will have exited.\n"
                    "You can run `remi desktop attach-session` to re-attach to the session."
                )
                cmd_suffix += '; exec bash'

            # Keep the session running only if the command fails
            else:
                cmd_suffix += ' || exec bash'

            cmd_suffix = '\"' + cmd_suffix + '\"'

        # If not in background, the singularity cmd is appended as a list of arguments.
        elif use_container:
            cmd += singularity_cmd

        cmd.append(cmd_suffix)

    run_remote_command_inria(
        command=cmd,
        hostname=hostname,
        check=True,
        force_tty_allocation=True,
        enable_x_forwarding=enable_x_forwarding
    )

    if background and attach:
        _attach_session(
            hostname=hostname,
            session_name=session_name
        )


def _attach_session(
    hostname: str,
    session_name: str
) -> None:

    background_backend: str = Config().desktop.background.backend

    cmd: list[str]
    if background_backend == 'screen':
        cmd = ['screen', '-r', session_name]
    elif background_backend == 'tmux':
        cmd = ['tmux', 'attach-session', '-t', session_name]

    run_remote_command_inria(
        cmd,
        hostname=hostname,
        check=True,
        force_tty_allocation=True
    )


#######
# CLI #
#######
_option_attach: click.option = click.option(
    '-a', '--attach',
    default=False,
    is_flag=True,
    help="Automatically attach to screen/tmux session (Only useful when `background` is enabled."
)

# pylint: disable=unused-argument
def _callback_option_session_name(ctx, param, value) -> str:
    # Make sure the script is a valid file
    if value:
        return value

    return Config().project_name


_option_session_name: click.option = click.option(
    '-n', '--name',
    'session_name',
    type=str,
    callback=_callback_option_session_name,
    help="Manually name the screen/tmux session"
)

@click.group(cls=OrderedGroup, invoke_without_command=True)
@click.pass_context
@option_script
@option_hostname
@option_background
@_option_session_name
@_option_attach
@option_xforwarding
@option_no_push
@option_no_build
def desktop(
    ctx,
    script_name: str,
    background: bool,
    session_name: str,
    attach: bool,
    hostname: str,
    x_forwarding: bool,
    no_push: bool,
    no_build: bool
) -> None:
    """
    Run a script on the remote workstation.
    """
    if ctx.invoked_subcommand is None:
        ctx.invoke(
            desktop_script,
            script_name=script_name,
            hostname=hostname,
            background=background,
            session_name=session_name,
            attach=attach,
            x_forwarding=x_forwarding,
            no_push=no_push,
            no_build=no_build
        )


@desktop.command('script')
@option_script
@option_hostname
@option_background
@_option_session_name
@_option_attach
@option_xforwarding
@option_no_push
@option_no_build
def desktop_script(
    script_name: str,
    hostname: str,
    background: bool,
    session_name: str,
    attach: bool,
    x_forwarding: bool,
    no_push: bool,
    no_build
) -> None:

    _desktop(
        sub_command='script',
        script_name=script_name,
        hostname=hostname,
        background=background,
        session_name=session_name,
        attach=attach,
        enable_x_forwarding=x_forwarding,
        no_push=no_push,
        no_build=no_build
    )


@desktop.command('command')
@option_hostname
@option_background
@_option_session_name
@_option_attach
@option_xforwarding
@option_no_push
@option_no_build
@argument_command
def desktop_command(
    command: list[str],
    background: bool,
    session_name: str,
    attach: bool,
    hostname: str,
    x_forwarding: bool,
    no_push: bool,
    no_build: bool
) -> None:

    _desktop(
        sub_command='command',
        command=command,
        background=background,
        session_name=session_name,
        attach=attach,
        hostname=hostname,
        enable_x_forwarding=x_forwarding,
        no_push=no_push,
        no_build=no_build
    )


@desktop.command('interactive')
@option_hostname
@option_xforwarding
@option_no_push
@option_no_build
def desktop_interactive(
    hostname: str,
    x_forwarding: bool,
    no_push: bool,
    no_build: bool
) -> None:

    _desktop(
        sub_command='interactive',
        hostname=hostname,
        enable_x_forwarding=x_forwarding,
        no_push=no_push,
        no_build=no_build
    )


argument_session_name: click.argument = click.argument(
    'session_name',
    default='',
    type=str,
    callback=_callback_option_session_name,
    nargs=1
)

@desktop.command('attach-session')
@argument_session_name
@option_hostname
def desktop_attach_session(
    hostname: str,
    session_name: str
) -> None:
    _attach_session(
        hostname=hostname,
        session_name=session_name
    )



"""
Command line interface for remi.
"""

import sys
from os.path import isdir
import click

from .user_interaction import print_error
from .cli_utils import OrderedGroup

from . import (setup_commands,
               file_transfer,
               remote_compute,
               remote_servers)


@click.group(cls=OrderedGroup)
@click.pass_context
def remi(ctx) -> None:
    """
    RemI: Remote Inria.

    Documentation:  https://remote-inria.gitlabpages.inria.fr/
    Source code:    https://gitlab.inria.fr/remote-inria/remi
    """
    if ctx.invoked_subcommand != 'init':
        if not isdir('.remi/'):
            print_error(
                "RemI directory (`.remi/`) was not found ! Exiting."
                "\nYou might want to `cd` to your project or to run"
                " `remi init` if you have not done it yet."
            )
            sys.exit(1)


# Setup
remi.add_command(setup_commands.init)
remi.add_command(setup_commands.setup)
remi.add_command(setup_commands.update)

# File transfer
remi.add_command(file_transfer.clean)
remi.add_command(file_transfer.pull)
remi.add_command(file_transfer.push)

# Remote desktop
remi.add_command(remote_compute.desktop.desktop)

# Cluster
remi.add_command(remote_compute.singularity.build_container)
remi.add_command(remote_compute.clusters.inria_cluster.cluster)

# GriCAD
remi.add_command(remote_compute.clusters.gricad.gricad)

# SSH tunnel
remi.add_command(remote_servers.remote_server.ssh_bridge)

# Jupyter notebook
remi.add_command(remote_servers.jupyter.jupyter)

# TensorBoard
remi.add_command(remote_servers.tensorboard.tensorboard)

# Aim
remi.add_command(remote_servers.aim.aim)



"""
Common options for remote servers commands
"""

import click

from ..config import Config


def option_port(service_name: str):

    # pylint: disable=unused-argument
    def _callback(ctx, param, value) -> int:
        if value is not None:
            return value

        # The default port for an SSH tunnel is the TensorBoard port.
        if service_name == 'ssh tunnel':
            _service_name = 'tensorboard'
        else:
            _service_name = service_name

        return getattr(Config().remote_servers, _service_name).port

    return click.option(
        '-p', '--port',
        type=int,
        callback=_callback,
        help=f"Local port for accessing {service_name}."
    )


# pylint: disable=unused-argument
def _callback_option_browser(ctx, param, value) -> bool:
    if value is not None:
        return value

    config: Config = Config()

    if ctx.command.name == 'jupyter':
        return config.jupyter.open_browser

    if ctx.command.name == 'tensorboard':
        return config.tensorboard.open_browser

    return True


option_open_browser = click.option(
    '--browser/--no-browser',
    'open_browser',
    default=True,
    callback=_callback_option_browser,
    help="Open the local browser after running server."
)



"""
CLI utilities for the cluster commands.
"""

import click

from ...config import Config


argument_oar_job_id: click.argument = click.argument(
    'oar-job-id',
    type=int,
    nargs=1
)

argument_oar_job_id_list: click.argument = click.argument(
    'oar-job-ids',
    type=int,
    nargs=-1
)



"""
Aim support.
https://github.com/aimhubio/aim
"""

import click

from .remote_server import RemoteServer
from ..cli_utils import OrderedGroup, option_hostname
from .cli_utils import option_port, option_open_browser
from ..config import Config


class Aim(RemoteServer):
    """
    Aim remote server.
    https://github.com/aimhubio/aim
    """

    def __init__(
            self,
            port: int,
            hostname: str,
            repo: str = '',
            open_browser: bool = True
    ) -> None:

        super().__init__(
            server_start_command=f"aim up --port {port} --repo {repo}",
            server_stop_command="pkill aim && pkill uvicorn",
            port=port,
            hostname=hostname,
            open_browser=open_browser,
            local_url=f'http://localhost:{port}'
        )

    def _server_start_callback(self, server_output_line: str) -> bool:
        return server_output_line.startswith('Open ')


#######
# CLI #
#######
_option_aim_port = option_port(service_name='aim')


# pylint: disable=unused-argument
def _callback_option_aim_repo(ctx, param, value) -> str:
    if value:
        return value

    return Config().remote_servers.aim.repo


_option_aim_repo = click.option(
    '-r', '--repo',
    callback=_callback_option_aim_repo,
    type=str,
    help="Log directory."
)


@click.group(cls=OrderedGroup, invoke_without_command=True)
@click.pass_context
@_option_aim_port
@option_hostname
@_option_aim_repo
@option_open_browser
def aim(
        ctx,
        port: int,
        hostname: str,
        repo: str,
        open_browser: bool
) -> None:
    """
    Run Aim on the remote host.
    """
    if ctx.invoked_subcommand is None:
        ctx.invoke(
            aim_run,
            port=port,
            hostname=hostname,
            repo=repo,
            open_browser=open_browser
        )


@aim.command('run')
@_option_aim_port
@option_hostname
@option_open_browser
@_option_aim_repo
def aim_run(
        port: int,
        hostname: str,
        repo: str,
        open_browser: bool
) -> None:
    """
    Run Aim on the remote host.
    """
    Aim(
        hostname=hostname,
        port=port,
        repo=repo,
        open_browser=open_browser
    ).start()


@aim.command('stop')
@_option_aim_port
@option_hostname
def aim_stop(port: int, hostname: str) -> None:
    """
    Stop Aim on the remote host.
    """
    Aim(
        hostname=hostname,
        port=port
    ).stop_server()
